import numpy as np
import cv2 as cv


def opticalFlow(
    videoName, videoFps, videoStartIndex, numOfFramesToSkip, distancePerPx, orientation
):
    # Open capture of video at the set start index
    cap = cv.VideoCapture("./Resources/" + videoName)
    cap.set(cv.CAP_PROP_POS_FRAMES, videoStartIndex)

    # Parameteres for Lucas-Kanade optical flow
    lucas_kanade = dict(
        winSize=(15, 15),
        maxLevel=2,
        criteria=(cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT, 10, 0.03),
    )
    # Parameters for shiTomasi corner detection
    features = dict(maxCorners=500, qualityLevel=0.3, minDistance=7, blockSize=7)

    # Save the frame index of the last estimation to keep track of elapsed  time
    frameIndexLastEstimation = -1

    track_length = 10
    detect_interval = 5
    tracks = []
    DECIMAL = 4
    frameIndex = 0
    speedY = 0
    speedX = 0

    # Save speed at each estimation for data analysis
    speedData = []

    while True:
        # Read frame from capture
        ret, frame = cap.read()

        # Ret will be False if there are no more frames in the video
        if not ret:
            return speedData

        # Convert frame to grayscale
        gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)

        # If tracks have been made
        if len(tracks) > 0:
            image0, image1 = previous_gray, gray

            # Get initial points from tracks
            p0 = np.float32([tr[-1] for tr in tracks]).reshape(-1, 1, 2)

            # Find new points with optical flow
            p1, st, err = cv.calcOpticalFlowPyrLK(
                image0, image1, p0, None, **lucas_kanade
            )

            # Find p0 reversed, points in the other direction
            p0r, st, err = cv.calcOpticalFlowPyrLK(
                image1, image0, p1, None, **lucas_kanade
            )

            # Find the difference between p0 and p0r to make sure p1 is a good match
            difference = abs(p0 - p0r).reshape(-1, 2).max(-1)

            # Good matches between p0 and p1
            good = difference < 1

            new_tracks = []

            # Go through all new points
            for tr, (x, y), good_flag in zip(tracks, p1.reshape(-1, 2), good):
                # Points not  a good match, continue...
                if not good_flag:
                    continue

                # Add points to track
                tr.append((x, y))

                # Delete the first point if the track is too long.
                if len(tr) > track_length:
                    del tr[0]

                new_tracks.append(tr)
            tracks = new_tracks

        # Update points every set interval of frames
        if frameIndex % detect_interval == 0:
            mask = np.zeros_like(gray)

            # Set mask
            mask[100:200] = 255

            # Find the corners in the frame to track
            p = cv.goodFeaturesToTrack(gray, mask=mask, **features)

            # If points are found
            if p is not None:
                # Reshape points and add to tracks
                for x, y in np.float32(p).reshape(-1, 2):
                    tracks.append([(x, y)])

        # self.tracks[a][b][c]
        # a is for the main tracking points
        # b is 10 for the trail of the point
        # c is the x or y value of the point
        totalDistanceX = 0
        totalDistanceY = 0
        speedX = 0
        speedY = 0

        # Check if frame index is in interval of set frames between estimations
        if frameIndex % numOfFramesToSkip == 0:
            for a in range(len(tracks)):
                if len(tracks[a]) >= 10:
                    # Distance in X-direction
                    totalDistanceX = totalDistanceX + tracks[a][9][0] - tracks[a][0][0]

                    # Distance in Y-direction
                    totalDistanceY = totalDistanceY + tracks[a][9][1] - tracks[a][0][1]

            # Check if tracks is larger than 0
            if len(tracks) > 0:
                distanceX = totalDistanceX / len(tracks)
                distanceY = totalDistanceY / len(tracks)

                # Calculate how much time has elapsed since last estimation
                time = (frameIndex - frameIndexLastEstimation) / videoFps

                # Convert distance per pixles to distance in meters
                distanceInMetersX = distanceX * distancePerPx
                distanceInMetersY = distanceY * distancePerPx

                # Use the estimated distance to calculate speed in Y-axis
                speedX = distanceInMetersX / time
                speedY = distanceInMetersY / time

                # Update frame index
                frameIndexLastEstimation = frameIndex

                #   Add Y-speed to speedata for later analyzing
                speedData.append(speedY * -orientation)

            previous_gray = gray

        # Update previous frame and index
        previous_gray = gray
        frameIndex += 1