import inspect
import math
import pprint
import statistics
import sys
import time
import csv
import cv2 as cv
import os
import numpy as np

# Fix path when importing files from different folder
sys.path.append("./Algorithms")
from featureMatching import featureMatching
from opticalFlow import opticalFlow


# Convert time to a more readable format in ms
def convertTime(time):
    return round(time * 1000, 2)


def calculateRMSE(estimates, solution):
    # Create an array with referenceSpeeds
    targets = np.full((1, len(estimates)), solution)

    return np.sqrt(((estimates - targets) ** 2).mean())


# Write data to a .csv file for later analyzing
def writeDataToFile(data):
    # Go through every algorithm in data dictionary
    for algorithm, algorithmData in data.items():
        # Go through every general video
        for videos, videosData in algorithmData.items():
            filePath = "./Testing/Results/" + "/" + videos

            # Create empty folder if it does not exist
            try:
                os.mkdir(filePath)
            except FileExistsError as err:
                # Folder already exists, continue...
                pass

            # Create own results file for each algorithm
            with open(
                filePath + "/" + videos + "_" + algorithm + ".csv", "w", newline=""
            ) as file:
                # Open CSV writer with delimiter (value seperator) as ;
                writer = csv.writer(file, delimiter=";")

                # Get value names from dictionary
                valueNames = list(videosData.values())[0].keys()

                # Each reference speed should be on the first row
                referenceSpeeds = []

                # Save all reference speeds
                for videoData in videosData.values():
                    referenceSpeeds.append(str(videoData["Hastighet referanse (m/s)"]))

                # Write the first row to file with videos name
                writer.writerow([videos])

                # Write second row to file with reference speeds
                writer.writerow(["Referansehastighet (m/s)", *referenceSpeeds])

                # Go through every video and write data to respective rows
                for value in valueNames:
                    rowValues = []

                    # Get values from videoes
                    for videoData in videosData.values():
                        # Convert value to comma seperated string
                        # Later analyzing is done in google sheets which does not like dot seperation between integers and decimals
                        dataValue = str(format(videoData[value], ".2f"))

                        # print(value)
                        dataValue = dataValue.replace(".", ",")

                        # Save row value
                        rowValues.append(dataValue)

                    # Write row to file
                    writer.writerow([value, *rowValues])


VIDEO_FPS = 30
VIDEO_START_INDEX = 0
NUM_OF_FRAMES_TO_SKIP = 10

# 3cm distance between squares in net
DISTANCE_BETWEEN_SQUARES = 0.03

# -- Simulation video ---
NUM_OF_SQUARES_IN_FRAME_SIM = 15
VIDEO_HEIGHT_SIM = 720
DISTANCE_PER_PX_SIM = (0.10 * NUM_OF_SQUARES_IN_FRAME_SIM) / VIDEO_HEIGHT_SIM

# -- Remora video ---
NUM_OF_SQUARES_IN_FRAME_REM = 5
VIDEO_HEIGHT_REM = 480
DISTANCE_PER_PX_REM = (
    DISTANCE_BETWEEN_SQUARES * NUM_OF_SQUARES_IN_FRAME_REM
) / VIDEO_HEIGHT_REM

print(DISTANCE_PER_PX_SIM)
print(DISTANCE_PER_PX_REM)

# Constants
VIDEO_DATA = {
    "sim_realtime": {
        "distancePerPx": DISTANCE_PER_PX_SIM,
        "orientation": 1,
        "videos": {
            1: {"videoName": "sim_realtime_0.25x.mp4", "referenceSpeed": 0.05},
            2: {"videoName": "sim_realtime_0.5x.mp4", "referenceSpeed": 0.1},
            3: {"videoName": "sim_realtime.mp4", "referenceSpeed": 0.2},
            4: {"videoName": "sim_realtime_2x.mp4", "referenceSpeed": 0.4},
            1: {"videoName": "sim_realtime_4x.mp4", "referenceSpeed": 0.8},
        },
    },
    "remora_new": {
        "distancePerPx": DISTANCE_PER_PX_REM,
        "orientation": -1,
        "videos": {
            1: {"videoName": "remora_transformed_0.25x.avi", "referenceSpeed": 0.05},
            2: {"videoName": "remora_transformed_0.5x.avi", "referenceSpeed": 0.1},
            1: {"videoName": "remora_transformed.avi", "referenceSpeed": 0.2},
            4: {"videoName": "remora_transformed_2x.avi", "referenceSpeed": 0.4},
            5: {"videoName": "remora_transformed_4x.avi", "referenceSpeed": 0.8},
        },
    },
}

ALGORITHMS = [featureMatching, opticalFlow]


data = {}


def runAlgorithm(
    algorithm, distancePerPx, orientation, video, videoName, referenceSpeed
):
    # Check if argument is an algorithm
    if not inspect.isfunction(algorithm):
        raise ValueError("Argument is not a function")

    # Run the algorithm
    startTime = time.time()
    speedData = algorithm(
        videoName,
        VIDEO_FPS,
        VIDEO_START_INDEX,
        NUM_OF_FRAMES_TO_SKIP,
        distancePerPx,
        orientation,
    )
    endTime = time.time()
    # Calculate how much time the algorithm used to run
    timeUsed = endTime - startTime

    # Convert variables to a more readable format
    timeUsed = convertTime(timeUsed)

    # If there is no data in speedData variable
    if speedData is None or len(speedData) == 0:
        # Add algorithm data to data dictionary with 0 values
        data[algorithm.__name__][videoName] = {
            "Tid total": timeUsed,
            "Hastighet gjennomsnitt": 0,
            "Hastighet min": 0,
            "Hastighet max": 0,
            "Hastighet varianse": 0,
            "Hastighet standardavvik": 0,
            "Relativ feil": 0,
            "Nøyaktighet": 0,
            "Hastighet referanse": referenceSpeed,
            "Antall bilder per estimering": 0,
            "Presisjon": 0,
            "RMSE": 0,
        }
        print(algorithm.__name__ + " " + videoName + " no matches found")

        # No need to run the rest of the code in function
        return

    # Open video capture to find number of frames in video
    cap = cv.VideoCapture("./Resources/" + videoName)
    numOfFramesInVideo = int(cap.get(cv.CAP_PROP_FRAME_COUNT))
    numOfEstimations = len(speedData) or 1
    numOfFramesPerEstimation = math.ceil(numOfFramesInVideo / numOfEstimations)

    # Find the average speed
    totalSpeed = 0
    for s in speedData:
        totalSpeed += abs(s)
    if len(speedData) == 0:
        speedAvg = 0
    else:
        speedAvg = totalSpeed / len(speedData)

    # Other speed statistical calculations
    speedMin = min(speedData)
    speedMax = max(speedData)
    speedVar = statistics.variance(speedData)
    speedDev = statistics.pstdev(speedData)
    precision = speedMax - speedMin

    relPrecision = 0
    if speedMax - speedMin != 0:
        relPrecision = speedAvg / (speedMax - speedMin)

    rmse = calculateRMSE(speedData, referenceSpeed)

    # Calculate relative error with reference value and average
    relError = (abs(speedAvg - referenceSpeed)) / abs(referenceSpeed)

    # Calculate accuracy error with reference value and average
    accuracy = (
        referenceSpeed - (max(referenceSpeed, speedAvg) - min(referenceSpeed, speedAvg))
    ) / referenceSpeed

    # If the calculated accuracy is outside of the 0-100% range of the reference value
    # the accuracy must be 0
    if accuracy < 0:
        accuracy = 0

    # Convert data to two decimals points
    speedAvg = round(speedAvg, 2)
    referenceSpeed = round(referenceSpeed, 2)
    speedMin = round(speedMin, 2)
    speedMax = round(speedMax, 2)
    speedVar = round(speedVar, 2)
    speedDev = round(speedDev, 2)
    relError = round(relError * 100, 2)
    accuracy = round(accuracy * 100, 2)
    numOfFramesPerEstimation = int(numOfFramesPerEstimation)
    precision = round(precision, 2)
    relPrecision = round(relPrecision * 100, 2)
    rmse = round(rmse, 2)

    # Add algorithm data to data dictionary
    data[algorithm.__name__][video][videoName] = {
        "Tid total (ms)": timeUsed,
        "Hastighet gjennomsnitt (m/s)": speedAvg,
        "Hastighet min (m/s)": speedMin,
        "Hastighet max (m/s)": speedMax,
        "Hastighet varianse (m/s)^2": speedVar,
        "Hastighet standardavvik (m/s)": speedDev,
        "Relativ feil (%)": relError,
        "Nøyaktighet (%)": accuracy,
        "Hastighet referanse (m/s)": referenceSpeed,
        "Antall bilder per estimering": numOfFramesPerEstimation,
        "Presisjon": precision,
        "Relativ resisjon (%)": relPrecision,
        "RMSE": rmse,
    }


def runAllAlgorithms():
    # Go thorugh every algorithm
    for a in ALGORITHMS:
        # Add empty dictionary to data array for algorithm
        data[a.__name__] = {}

        # Go through every general video
        for video in VIDEO_DATA:
            # Add empty dictionary to data array for video
            data[a.__name__][video] = {}
            distancePerPx = VIDEO_DATA[video]["distancePerPx"]
            orientation = VIDEO_DATA[video]["orientation"]

            # Go thorugh every specific video with different speeds
            for index, values in VIDEO_DATA[video]["videos"].items():
                videoName = values["videoName"]
                referenceSpeed = values["referenceSpeed"]

                # Run specific algorithm on the specific video
                runAlgorithm(
                    a,
                    distancePerPx,
                    orientation,
                    video,
                    videoName,
                    referenceSpeed,
                )

    # Save data to .CSV files for later analyzing
    writeDataToFile(data)


runAllAlgorithms()