import cv2 as cv
import numpy as np
import utils

# Equalize histogram to fill spectrum of luminosity
def flattenHistogram(img):
    # Convert BGR image to LAB image (Luminosity, A-channel, B-channel)
    img_lab = cv.cvtColor(img, cv.COLOR_BGR2LAB)

    # Split channels
    l, a, b = cv.split(img_lab)

    # Old method
    # equalized = cv.equalizeHist(l)
    # merged_img = cv.merge((equalized, a, b))

    # CLAHE algorithm
    clahe = cv.createCLAHE(clipLimit=3.0, tileGridSize=(8, 8))
    clahe_img = clahe.apply(l)

    merged_img = cv.merge((clahe_img, a, b))

    return cv.cvtColor(merged_img, cv.COLOR_LAB2BGR)


# Detect edges using canny edge detection
# Parameters for edge detection is automatically optimized based
# on lighting conditions, contrast etc.
def cannyEdgeDetecion(img, sigma=0.33):
    # Find the median of the all pixel intensities in the image
    # median = np.median(img)

    # Automaticallly choose lower and upper boundries based on median
    # lower = int(max(0, (1.0 - sigma) * median))
    # upper = int(max(255, (1.0 + sigma) * median))

    # Convert to grayscale
    gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)

    median = cv.medianBlur(gray, 3)

    # Apply threshold
    thresh = cv.threshold(median, 150, 255, cv.THRESH_BINARY)[1]

    # Dilated mask - mask = morphology edgeout
    kernel = cv.getStructuringElement(cv.MORPH_RECT, (5, 5))
    dilate = cv.morphologyEx(thresh, cv.MORPH_DILATE, kernel)
    # dilate = cv.dilate(thresh, kernel, iterations=1)

    # Get absolute difference between dilaton and threshold
    diff = cv.absdiff(dilate, thresh)

    edges = 255 - diff

    # Blur the image using median blurring to remove noise
    # blurred_img =  cv.GaussianBlur(img, (3,3), 0)

    # Detect edges using Canny in OpenCV
    # edge = cv.Canny(thresh)

    return edges


# DOES THIS WORK????
def detectCorners(img, img2):
    copyImg = img.copy()
    # img_new = cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    # img_new = np.float32(img)
    corners = cv.goodFeaturesToTrack(copyImg, 200, 0.01, 10)

    if corners is not None:
        corners = np.int0(corners)

        for corner in corners:
            x, y = corner.ravel()
            cv.circle(img2, (x, y), 5, 255, -1)

    return img2


def findKeyPoints(frame, method="ORB", numOfKeypoints=100):
    # Check if method is a valid value
    methods = ["ORB", "SIFT", "FAST"]
    if method not in methods:
        raise ValueError("Invalid method type. Expected one of: %s" % methods)

    # Convert frame to grayscale
    frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)
    # frame_gray = cannyEdgeDetecion(frame)

    # ORB - Oriented FAST and Rotated BRIEF
    if method == methods[0]:
        orb = cv.ORB_create(nfeatures=numOfKeypoints, patchSize=15)
        kp, des = orb.detectAndCompute(frame_gray, None)

        # Convert to CV_32F format
        des = np.float32(des)
        return kp, des

    # SIFT - Scale-Invariant Feature Transform
    elif method == methods[1]:
        sift = cv.xfeatures2d.SIFT_create()
        return sift.detectAndCompute(frame_gray, None)

    # FAST for keypoints and BRIEF for desciptors
    elif method == methods[2]:
        # Find keypoints with FAST
        fast = cv.FastFeatureDetector_create()
        kp = fast.detect(frame_gray, None)

        # Find descriptors with BRIEF
        brief = cv.xfeatures2d.BriefDescriptorExtractor_create()
        kp, des = brief.compute(frame_gray, kp)
        return kp, des


# Find matches between two datasets of keypoints and descriptors for two images
def findMatches(img1, img2, kp1, kp2, des1, des2):
    # FLANN matcher parameters
    FLANN_INDEX_KDTREE = (
        0  # Index for KDTREE algorithm used with floating point descriptors
    )
    FLANN_INDEX_LSH = 1  # Index for LSH algorithm used with binary desciptors
    RATIO_THRESHOLD = 0.7  # Threshold ratio value for a good match

    index_params = dict(
        algorithm=FLANN_INDEX_LSH,
        table_number=12,
        key_size=20,
        multi_probe_level=1,
    )
    search_params = dict(checks=50)

    # Create a FLANN matcher object and use knnMatch to find matches
    flann = cv.FlannBasedMatcher(index_params, search_params)
    matches = flann.knnMatch(des1, des2, k=2)

    # Save the good matches in a new array
    good_matches = []

    # Lowe's ratio test for good matches
    for m1, m2 in matches:
        # If the matches are far enough from each other, they are considered a good match
        if m1.distance < RATIO_THRESHOLD * m2.distance:
            good_matches.append([m1])

    return good_matches


# Filter the matches based on direction
def filterMatches(matches, kp1, kp2):
    # Keypoints must be of same length
    if len(kp1) != len(kp2):
        raise ValueError("Keypoint arrays are not of the same length")

    # Filter out the best n-matches
    NUM_OF_MATCHES = 5
    matches = matches[-NUM_OF_MATCHES:]

    # Minimum distance for a match to pass the filter
    DISTANCE_THRESHOLD = 150

    matchesUpwards = []
    matchesDownwards = []

    for m in matches:
        # Index of keypoints to match object
        index1 = m[0].queryIdx
        index2 = m[0].trainIdx

        # Find the distance between the keypoints in the Y-direction
        distance = kp1[index2].pt[1] - kp2[index1].pt[1]

        # Save match based on distance which determines the direction
        if distance > 0:
            matchesUpwards.append(m)
        else:
            matchesDownwards.append(m)

    return matchesDownwards, matchesUpwards


# Find the distance between two keypoints in a match
def distanceBetweenMatches(matches, kp1, kp2):
    if len(matches) == 0:
        return 0, 0

    # Save distance as average in X- and Y-direction
    avgDistanceY = 0
    avgDistanceX = 0

    for m in matches:
        # Get indexes for keypoints from match object
        index1 = m[0].queryIdx
        index2 = m[0].trainIdx

        # Calculate distances between keypoints based on index
        avgDistanceX += kp1[index1].pt[0] - kp2[index2].pt[0]
        avgDistanceY += kp1[index1].pt[1] - kp2[index2].pt[1]

    avgDistanceY = avgDistanceY / len(matches)
    avgDistanceX = avgDistanceX / len(matches)

    return avgDistanceX, avgDistanceY
