import numpy as np
import cv2 as cv
import sys
import time

import multiprocessing as mp
import ctypes as c
from os import system, name

# Fix path when importing files from different folder
sys.path.append("./")
from utils import rescaleFrame
from image import (
    cannyEdgeDetecion,
    findKeyPoints,
    findMatches,
    distanceBetweenMatches,
    filterMatches,
)
from measurements import calculateSpeed, calculatePosition


# CONSTANTS
KEYPOINTS_METHOD = "ORB"
VIDEO_NAME = "./Resources/sim_realtime.mp4"
# Height, width of the frames
VIDEO_RESOLUTION = (1080, 1920)
FRAME_START_INDEX = 0


def clear():
    # Clear console log in windows
    if name == "nt":
        _ = system("cls")


def printTime(times):
    # Clear previous outputs
    clear()

    for t in times:
        print(t[0], convertTime(t[1]))


def convertTime(time):
    return str(round(time * 1000, 2)) + "ms"


def updateTime(timePrev, timeBefore, timeAfter):
    timeNew = abs(timeAfter - timeBefore)
    return (timePrev + timeNew) / 2


cap = cv.VideoCapture(VIDEO_NAME)
cap.set(cv.CAP_PROP_POS_FRAMES, FRAME_START_INDEX)

matchesBetweenFrames = []


def computeSpeedAndDistance(running, frameIndex, speedX, speedY, x, y):
    # Local variables
    speedPrevX, speedPrevY = 0, 0
    kp, des = None, None
    kpTime = 0
    matchTime = 0
    previousIndex = 0

    # It's own capture reading from video to avoid race conditions in reading
    cap = cv.VideoCapture(VIDEO_NAME)
    cap.set(cv.CAP_PROP_POS_FRAMES, FRAME_START_INDEX)

    # Initial frame reading
    _, framePrev = cap.read()
    # Initial previous keypoints and descriptors
    kpPrev, desPrev = findKeyPoints(framePrev, KEYPOINTS_METHOD, 100)

    averageSpeed = 0

    while running.value == 1:
        frameIndexCurrent = frameIndex.value
        cap.set(cv.CAP_PROP_POS_FRAMES, frameIndexCurrent + FRAME_START_INDEX)

        numOfFramesBetweenEstimation = frameIndexCurrent - previousIndex

        ret, frame = cap.read()

        if not ret:
            break

        # Calculate keypoints and descriptors for the current frame
        timeBeforeKp = time.time()
        kp, des = findKeyPoints(frame, KEYPOINTS_METHOD, 100)
        timeAfterKp = time.time()

        # Find feature matches between previous and current frame
        timeBeforeMatching = time.time()
        matchesBetweenFrames = findMatches(frame, framePrev, kp, kpPrev, des, desPrev)
        timeAfterMatching = time.time()

        matchesDownwards, matchesUpwards = filterMatches(
            matchesBetweenFrames, kp, kpPrev
        )

        # Too many frames is skipped
        if timeAfterMatching - timeBeforeMatching > 300:
            print("Skipping...")
            continue

        # Calculate the average distance between good matches
        distanceX, distanceY = distanceBetweenMatches(matchesUpwards, kp, kpPrev)

        # Use the estimated distance to calculate speed in X-axis
        newSpeedX, speedPrevX = calculateSpeed(
            distanceX, speedPrevX, numOfFramesBetweenEstimation
        )

        # Use the estimated distance to calculate speed in Y-axis
        newSpeedY, speedPrevY = calculateSpeed(
            distanceY, speedPrevY, numOfFramesBetweenEstimation
        )

        # Use the estimated speed to calculate an updated position
        newX, newY = calculatePosition(
            x.value, y.value, newSpeedX, newSpeedY, numOfFramesBetweenEstimation
        )

        # Update global shared variables
        speedX.value = newSpeedX
        speedY.value = newSpeedY
        x.value = newX
        y.value = newY

        averageSpeed += newSpeedY
        averageSpeed = averageSpeed / 2

        print("Average speed: ", averageSpeed)

        # Performance/time statistics
        kpTime = updateTime(kpTime, timeBeforeKp, timeAfterKp)
        matchTime = updateTime(matchTime, timeBeforeMatching, timeAfterMatching)
        # print(printTime([["Keypoints time:", kpTime], ["Matching time:", matchTime]]))

        # Make copies of the frames to avoid mutating frames
        # (current frame is going to be used in the next estimation as previous frame)
        frameCopy = frame.copy()
        framePrevCopy = framePrev.copy()

        # Add text to show which frame is which
        cv.putText(
            frameCopy,
            "Current frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )
        cv.putText(
            frameCopy,
            "Frame index: " + str(frameIndexCurrent),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Frame index: " + str(previousIndex),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Previous frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        # Need to draw only good matches, so create a mask
        matchesMask = [[1, 0] for i in range(len(matchesBetweenFrames))]

        draw_params = dict(
            matchColor=(0, 255, 0),
            singlePointColor=(0, 0, 255),
            matchesMask=matchesMask,
            flags=0,
        )

        # Need to draw only good matches, so create a mask
        matchesMaskFiltered = [[1, 0] for i in range(len(matchesUpwards))]

        # ratio test as per Lowe's paper
        # for i, (m) in enumerate(matchesFiltered):
        #     if m.distance < 0.7 * n.distance:
        #         matchesMask[i] = [1, 0]

        draw_params_filtered = dict(
            matchColor=(0, 255, 0),
            singlePointColor=(0, 0, 255),
            # matchesMask=matchesMaskFiltered,
            flags=0,
        )

        matchesFrame = cv.drawMatchesKnn(
            framePrevCopy,
            kpPrev,
            frameCopy,
            kp,
            matchesBetweenFrames,
            None,
            **draw_params
        )

        matchesFrameFiltered = cv.drawMatchesKnn(
            framePrevCopy,
            kpPrev,
            frameCopy,
            kp,
            matchesUpwards,
            None,
            **draw_params_filtered
        )

        cv.imshow("matches with drawing", rescaleFrame(matchesFrame, 0.4))
        cv.imshow("matches filtered", rescaleFrame(matchesFrameFiltered, 0.4))

        k = cv.waitKey(30) & 0xFF

        # Show the Output
        if k == 27:
            running.value = 0
            break

        framePrev = frame
        kpPrev, desPrev = kp, des
        previousIndex = frameIndexCurrent


def drawFrames(running, frameIndex, speedX, speedY, x, y):
    _, framePrev = cap.read()
    kpPrev, desPrev = findKeyPoints(framePrev, KEYPOINTS_METHOD)

    kpTime = 0
    matchTime = 0
    drawingTime = 0
    totalPerFrame = 0

    while True:
        _, frame = cap.read()

        matchesFrame = frame

        frameCopy = frame.copy()
        framePrevCopy = framePrev.copy()

        # ---------------------------------------
        # -------- DRAWING AND RENDERING --------
        # ---------------------------------------

        # Add text to show which frame is which
        cv.putText(
            frameCopy,
            "Current frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )
        cv.putText(
            frameCopy,
            "Frame index: " + str(frameIndex.value),
            (50, 90),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            framePrevCopy,
            "Previous frame",
            (50, 50),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (255, 255, 255),
            2,
        )

        cv.putText(
            frameCopy,
            "SpeedY: {0:.2g}".format(speedY.value) + "m/s",
            (210, 225),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (0, 255, 255),
            2,
        )

        cv.putText(
            frameCopy,
            "Position: X{0:.2g}".format(x.value) + "m, Y{0:.2g}".format(y.value) + "m",
            (210, 265),
            cv.FONT_HERSHEY_TRIPLEX,
            1.0,
            (0, 255, 255),
            2,
        )

        frameIndex.value = frameIndex.value + 1

        cv.imshow("matches", rescaleFrame(frameCopy, 0.5))

        k = cv.waitKey(5) & 0xFF

        # Show the Output
        if k == 27:
            running.value = 0
            break


if __name__ == "__main__":
    # -- Shared variables between processes --

    # If True (1), all processes will terminate
    running = mp.Value("i", 1)

    # Keep track of the current frame index running in the video/camera feed
    frameIndex = mp.Value("i", 0)

    # Speed in X- and Y-axis of the robot
    speedX = mp.Value("d", 0)
    speedY = mp.Value("d", 0)

    # X- and Y-position of the robot
    x = mp.Value("d", 0)
    y = mp.Value("d", 0)

    processes = []
    # Tasks that should be run in parallell
    tasks = [drawFrames, computeSpeedAndDistance]

    # Run every task in parallell
    for t in tasks:
        # Create a process using the multiprocessing library
        # Notice the shared variables passed as arguments
        p = mp.Process(target=t, args=(running, frameIndex, speedX, speedY, x, y))

        # Save process in list
        processes.append(p)

        # Start running the process
        p.start()

    # Wait for every process to complete
    for p in processes:
        p.join()
