import numpy as np
import cv2 as cv
import sys

sys.path.append("./")
from measurements import calculateSpeed, calculatePosition


# Constants for selecting good points
# Points outside of the borders will not be used
BORDER_X = 200
BORDER_Y = 100
# HOT FIX for video with black bar on the RIGHT(!) side
# Remove later when changing video
BLACK_BAR_WIDTH = 320
# HOT FIX center border
BORDER_CENTER_Y1 = 360
BORDER_CENTER_Y2 = 440

# Find points to track using optical flow
def findPointsToTrack(frame, maxCorners=500):
    # Convert to grayscale
    frame = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)

    # Parameters for goodFeaturesToTrack method
    feature_params = dict(
        maxCorners=maxCorners, qualityLevel=0.1, minDistance=10, blockSize=5
    )

    # Use built-in method for finding features to track using Shi-Tomasi method'
    # This method will find the N strongest corners in the frame
    p0 = cv.goodFeaturesToTrack(frame, mask=None, **feature_params)

    # Store the new good points in a new array
    good_points = []

    # Filter points that is outside of the crop range specified
    for p in p0:
        if (
            p[0][0] - BORDER_X > 0
            and p[0][1] - BORDER_Y - 200 > 0
            and (p[0][1] < BORDER_CENTER_Y1 or p[0][1] > BORDER_CENTER_Y2)
            and p[0][0] + BORDER_X + BLACK_BAR_WIDTH < VIDEO_WIDTH
            and p[0][1] + BORDER_Y < VIDEO_HEIGHT
        ):
            good_points.append(p)
            print(p[0][1])

    # Points have to be of type float32 to be used with openCV methods
    p0 = np.float32(good_points)
    print(len(p0))

    return p0, frame


def updatePoints(p0, p1, frame, mask):
    MIN_NUM_OF_POINTS = 500
    p0New = []
    p1New = []
    maskNew = []

    # Make sure the arrays are of the correct shape
    p0 = np.float32(p0).reshape(-1, 1, 2)
    p1 = np.float32(p1).reshape(-1, 1, 2)

    # Go through all new points and check if any is outside of the set borders
    for i in range(len(p0)):
        if (
            p1[i][0][0] - BORDER_X > 0
            and p1[i][0][1] - BORDER_Y > 0
            and p1[i][0][0] + BORDER_X + BLACK_BAR_WIDTH <= VIDEO_WIDTH
            and p1[i][0][1] + BORDER_Y <= VIDEO_HEIGHT
        ):
            # Point is not outside of set border, add to new array
            p0New.append(p0[i])
            p1New.append(p1[i])
            maskNew.append(mask[i])

    # Check if there is enough points, if not add more
    if len(p1) < MIN_NUM_OF_POINTS:
        # Find new points based on number of missing points from constant
        newPoints, frame = findPointsToTrack(frame, MIN_NUM_OF_POINTS - len(p1))

        # Add new points to previous and current point-arrays
        for p in newPoints:
            p0New = np.append(p0New, p)
            p1New = np.append(p1New, p)

    # Make sure the new arrays are of the correct shape
    p0New = np.float32(p0New).reshape(-1, 1, 2)
    p1New = np.float32(p1New).reshape(-1, 1, 2)

    return p0New, p1New, frame_gray, mask


# Calculate the average distance between initialpoints and new points
def distanceBetweenPoints(initialPoints, points):
    totalDistanceX = 0
    totalDistanceY = 0
    numOfPointsX = 0
    numOfPointsY = 0
    avgDistanceX = 0
    avgDistanceY = 0

    THRESHOLD = 5

    # Go through all points to find total distance in X- and Y-direction
    for i in range(0, len(points)):
        distanceX = initialPoints[i][0][0] - points[i][0][0]
        distanceY = initialPoints[i][0][0] - points[i][0][1]

        if distanceX > THRESHOLD:
            numOfPointsX += 1
            totalDistanceX += distanceX

        if distanceY > THRESHOLD:
            numOfPointsY += 1
            totalDistanceY += distanceY
        else:
            print("Below THRESHOLD!!.......")

    if numOfPointsX == 0:
        avgDistanceX = 0
        print("No points in x direction")
    else:
        avgDistanceX = totalDistanceX / numOfPointsX

    if numOfPointsY == 0:
        print("No points in Y direction")
        avgDistanceY = 0
    else:
        avgDistanceY = totalDistanceY / numOfPointsY

    # The average distance is the total distance divided by number of points
    return (avgDistanceX, avgDistanceY)


FRAME_START_INDEX = 0

# Open video with capture
cap = cv.VideoCapture("./Resources/sim_realtime.mp4")
# Set start of the video to set starting frame
cap.set(cv.CAP_PROP_POS_FRAMES, FRAME_START_INDEX)

# Save width and height of the video (current start frame)
VIDEO_WIDTH = cap.get(cv.CAP_PROP_FRAME_WIDTH)
VIDEO_HEIGHT = cap.get(cv.CAP_PROP_FRAME_HEIGHT)

# Create some random colors for drawing purposes
color = np.random.randint(0, 255, (1000, 3))

# Get the first starting frame of the video
ret, old_frame = cap.read()

# Create a mask image for drawing purposes
mask = np.zeros_like(old_frame)

# Take first frame and find corners (good features/points to track) in it
p0, old_gray = findPointsToTrack(old_frame)

# Save points as initial for later usage
initialPoints = p0
frameIndex = 0
frameIndexLastEstimation = 0
speedX, speedY = 0, 0
speedPrevX, speedPrevY = 0, 0
x, y = 0, 0
averagespeed = 0

# paramters used for calcOpticalFlowPyrLK
lk_params = dict(
    winSize=(8, 8),
    maxLevel=2,
    criteria=(cv.TERM_CRITERIA_EPS | cv.TERM_CRITERIA_COUNT, 10, 0.03),
)

while cap.isOpened():
    # Get next frame from the capture video
    ret, frame = cap.read()

    # If the frame has not loaded correctly break the loop
    # Will happen after the last frame in video
    if not ret:
        print("Average speed: ", averagespeed)
        break

    # Convert frame to grayscale
    frame_gray = cv.cvtColor(frame, cv.COLOR_BGR2GRAY)

    # Calculate optical flow
    p1, st, err = cv.calcOpticalFlowPyrLK(old_gray, frame_gray, p0, None, **lk_params)

    # Select good points
    good_new = p1[st == 1].reshape(-1, 1, 2)
    good_old = p0[st == 1].reshape(-1, 1, 2)

    # Draw the tracks for optical flow tracking
    for i, (new, old) in enumerate(zip(good_new, good_old)):
        a, b = new.ravel()
        c, d = old.ravel()
        # mask = cv.line(mask, (a, b), (c, d), color[i].tolist(), 2)
        # frame = cv.circle(frame, (a, b), 5, color[i].tolist(), -1)
        mask = cv.line(mask, (a, b), (c, d), (0, 0, 255), 2)
        frame = cv.circle(frame, (a, b), 5, (0, 0, 255), -1)
    img = cv.add(frame, mask)

    cv.rectangle(
        img,
        (int(BORDER_X), int(BORDER_Y)),
        (int(VIDEO_WIDTH) - BORDER_X - BLACK_BAR_WIDTH, int(VIDEO_HEIGHT) - BORDER_Y),
        1,
    )
    cv.rectangle(
        img,
        (int(BORDER_X), int(BORDER_Y + 200)),
        (
            int(VIDEO_WIDTH) - BORDER_X - BLACK_BAR_WIDTH,
            int(VIDEO_HEIGHT) - BORDER_Y,
        ),
        (0, 255, 0),
        1,
    )
    cv.rectangle(
        img,
        (0, BORDER_CENTER_Y1),
        (int(VIDEO_WIDTH - BLACK_BAR_WIDTH), BORDER_CENTER_Y2),
        (0, 0, 255),
        1,
    )

    if frameIndex - frameIndexLastEstimation >= 10:
        # Find the avarage distance moved
        distanceX, distanceY = distanceBetweenPoints(good_old, good_new)

        # Calculate speed
        speedX, speedPrevX = calculateSpeed(
            distanceX, speedPrevX, frameIndex - frameIndexLastEstimation
        )

        speedY, speedPrevY = calculateSpeed(
            distanceY, speedPrevY, frameIndex - frameIndexLastEstimation
        )

        # Calculate position
        x, y = calculatePosition(
            x, y, speedX, speedY, frameIndex - frameIndexLastEstimation
        )

        averagespeed = (averagespeed + speedY) / 2

        frameIndexLastEstimation = frameIndex

    for i in range(0, len(p1)):
        if p1[i][0][0] < BORDER_X or p1[i][0][1] < BORDER_Y:

            # Check if new points are outside of borders in the frame
            good_old, good_new, frame_gray, mask = updatePoints(
                good_old, good_new, frame, mask
            )

            # Reset drawing mask
            mask = np.zeros_like(frame)

            print("Updating on frame index: ", frameIndex)

            # # Find the avarage distance moved
            # distanceX, distanceY = distanceBetweenPoints(good_old, good_new)

            # print("Distance y: ", distanceY)

            # # Calculate speed
            # speedX, speedPrevX = calculateSpeed(
            #     distanceX, speedPrevX, frameIndex - frameIndexLastEstimation
            # )

            # speedY, speedPrevY = calculateSpeed(
            #     distanceY, speedPrevY, frameIndex - frameIndexLastEstimation
            # )

            # # Calculate position
            # x, y = calculatePosition(
            #     x, y, speedX, speedY, frameIndex - frameIndexLastEstimation
            # )

            # frameIndexLastEstimation = frameIndex

            break

    # Now update the previous frame and previous points
    old_gray = frame_gray.copy()
    p0 = good_new.reshape(-1, 1, 2)

    # Add speed and positon text
    cv.putText(
        img,
        "Speed: {0:.2g}".format(speedY) + "m/s",
        (210, 180),
        cv.FONT_HERSHEY_TRIPLEX,
        1.0,
        (0, 255, 255),
        2,
    )

    cv.putText(
        img,
        "Position: X{0:.2g}".format(x) + "m, Y{0:.2g}".format(y) + "m",
        (210, 225),
        cv.FONT_HERSHEY_TRIPLEX,
        1.0,
        (0, 255, 255),
        2,
    )

    cv.imshow("frame", img)

    frameIndex += 1
    k = cv.waitKey(30) & 0xFF

    if k == 27:
        break


cv.destroyAllWindows()
cap.release()